#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <libnd.h>
#include <protocols/tcp/libnd_tcp.h>
#include <protocols/udp/libnd_udp.h>
#include <plugins/libnd_traceset.h>

#define LND_ANON_DEFAULT_SUFFIX ".anon"

static LND_Protocol *tcp;
static LND_Protocol *udp;

typedef struct lnd_anon {

  gboolean     crop_udp; /* Crops pcap caplen to beginning of UDP payload */
  gboolean     crop_tcp; /* Crops pcap caplen to beginning of TCP payload */
  
  char        *suffix;

} LND_Anon;

const char *
name(void)
{
  return "Anon";
}

const char *
author(void)
{
  return "Christian Kreibich <christian.kreibich@cl.cam.ac.uk>";
}

const char *
version(void)
{
  return "0.1";
}


gboolean
init(void)
{
  if (! libnd_plugin_find("Trace-Set"))
    {
      D(("Trace-Set plugin not found\n"));
      return FALSE;
    }
  
  if (! (tcp = libnd_proto_registry_find(LND_PROTO_LAYER_TRANS,
					 IPPROTO_TCP)))
    {
      D(("TCP protocol plugin not found.\n"));
      return FALSE;
    }

  if (! (udp = libnd_proto_registry_find(LND_PROTO_LAYER_TRANS,
					 IPPROTO_UDP)))
    {
      D(("UDP protocol plugin not found.\n"));
      return FALSE;
    }
  
  return TRUE;
}


static void
anon_help(void)
{
  printf("This plugin removes/anonymises various bits and pieces of given traces.\n"
	 "\n"
	 "USAGE: lndtool -r anon [--crop-udp|-U] [--crop-tcp|-T]\n"
	 "                       [--suffix|-s <STRING>]\n"
	 "                       [--names-file|-f <FILE>] [<TRACE2> <TRACE3> ...]\n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --names-file|-f FILE     Take filenames of traces from flatfile FILE.\n");
}


static gboolean
anon_crop_udp(LND_Packet *packet, pcap_dumper_t *dumper)
{
  guchar *udp_data = libnd_packet_get_data(packet, udp, 0);
  guchar *udp_header_end = udp_data + 8;
  struct pcap_pkthdr hdr = packet->ph;

  if (! udp_data)
    {
      D(("Couldn't obtain UDP data.\n"));
      return FALSE;
    }
  
  hdr.caplen = udp_header_end - packet->data;  
  pcap_dump((u_char*) dumper, &hdr, packet->data);
  
  return TRUE;
}


static gboolean
anon_crop_tcp(LND_Packet *packet, pcap_dumper_t *dumper)
{
  guchar *tcp_data = libnd_packet_get_data(packet, tcp, 0);
  guchar *tcp_header_end;
  struct tcphdr *tcphdr;
  struct pcap_pkthdr hdr = packet->ph;
  
  if (! libnd_tcp_get_headers(packet, NULL, &tcphdr))
    {
      D(("Couldn't obtain TCP header.\n"));
      return FALSE;      
    }
  
  tcp_header_end = tcp_data + libnd_tcp_get_header_length(tcphdr);
  hdr.caplen = tcp_header_end - packet->data;  
  pcap_dump((u_char*) dumper, &hdr, packet->data);

  return TRUE;
}


static gboolean
anon_trace(LND_TraceSet *set, LND_Trace *trace, LND_Anon *anon)
{
  LND_PacketIterator  pit;
  LND_Packet *packet;
  pcap_dumper_t *dumper;
  char dumper_name[MAXPATHLEN];

  g_snprintf(dumper_name, MAXPATHLEN, "%s%s",
	     libnd_trace_get_name(trace), anon->suffix);
  
  if (! (dumper = pcapnav_dump_open(libnd_trace_get_pcap_handle(trace),
				    dumper_name, PCAPNAV_DUMP_TRUNC)))
    {
      fprintf(stderr, "Could not create output file '%s'\n", dumper_name);
      return TRUE;
    }

  for (libnd_pit_init_mode(&pit, trace, LND_PACKET_IT_AREA_R);
       libnd_pit_get(&pit);
       libnd_pit_next(&pit))
    {
      packet = libnd_pit_get(&pit);      

      if (anon->crop_udp && libnd_packet_has_proto(packet, udp))
	{
	  if (anon_crop_udp(packet, dumper))
	    continue;
	}

      if (anon->crop_tcp && libnd_packet_has_proto(packet, tcp))
	{
	  if (anon_crop_tcp(packet, dumper))
	    continue;
	}
      
      /* Fallback -- if we get here just dump out unmodified. */
      pcap_dump((u_char*) dumper, &packet->ph, packet->data);
    }
  
  pcap_dump_close(dumper);
  return TRUE;
  TOUCH(set);
}


gboolean
run(LND_Trace *unused, LND_PluginArgs *args)
{
  LND_Anon anon;
  LND_TraceSet *traces;
  int i, files = 0;

  memset(&anon, 0, sizeof(LND_Anon));
  anon.suffix = LND_ANON_DEFAULT_SUFFIX;
  
  if (! (traces = libnd_traceset_new()))
    return FALSE;
  
  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  anon_help();
	  return FALSE;
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--names-file"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the name of a file that contains\n"
		     "a list of trace files, one filename per line, when\n"
		     "using the --names-file|-f option.\n");
	      return FALSE;
	    }
	  
	  if (! libnd_misc_can_read(args->argv[i]))
	    printf("Could not open '%s'\n", args->argv[i]);
	  else if (! libnd_traceset_add_trace_name_list(traces, args->argv[i]))
	    printf("Could not add all traces provided in '%s'\n", args->argv[i]);
	  else
	    files++;
	}
      else if (!strcmp(args->argv[i], "--crop-udp") ||
	       !strcmp(args->argv[i], "-U"))
	{
	  anon.crop_udp = TRUE;
	}
      else if (!strcmp(args->argv[i], "--crop-tcp") ||
	       !strcmp(args->argv[i], "-T"))
	{
	  anon.crop_tcp = TRUE;
	}
      else if (!strcmp(args->argv[i], "--suffix") ||
	       !strcmp(args->argv[i], "-s"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the suffix to be appended to each\n"
		     "anonymised trace to the --suffix|-s option.\n");
	      return FALSE;
	    }

	  anon.suffix = args->argv[i];
	}
      else if (args->argv[i][0] != '-') /* Often these are just mistyped args */
	{
	  if (! libnd_traceset_add_trace_name(traces, args->argv[i]))
	    printf("Trace file %s not added.\n", args->argv[i]);
	  else
	    files++;
	}
    }
  
  if (args->argc == 0)
    {
      anon_help();
      return FALSE;
    }

  if (files > 0)
    {
      libnd_traceset_foreach(traces, (LND_TraceSetCB) anon_trace, &anon);  
      libnd_traceset_free(traces);
    }

  return TRUE;
  TOUCH(unused);
}
