/*

Copyright (C) 2004 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "libnd_phdl.h"

static LND_PHDL_HeaderTempl *context_header = NULL;
static GList                *context_stack  = NULL;


LND_PHDL_HeaderTempl   *
libnd_phdl_ht_new(const char *header_name,
			    LND_ProtocolLayer layer,
			    guint64 magic)
{
  LND_PHDL_HeaderTempl *templ;

  if (!header_name || !*header_name)
    return NULL;

  if (! (templ = g_new0(LND_PHDL_HeaderTempl, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  templ->name  = g_strdup(header_name);
  templ->layer = layer;
  templ->magic = magic;

  if (! (templ->root  = libnd_phdl_ft_block_new("<root>")))
    {
      D(("Out of memory\n"));
      libnd_phdl_ht_free(templ);
      return NULL;
    }

  return templ;
}

void                    
libnd_phdl_ctxt_push(void *data)
{
  if (! data)
    return;

  context_stack = g_list_prepend(context_stack, data);
}


void *
libnd_phdl_ctxt_pop(void)
{
  void *result;

  if (! context_stack)
    return NULL;

  result = context_stack->data;
  context_stack = g_list_remove_link(context_stack,
				     context_stack);
  
  return result;
}


void *
libnd_phdl_ctxt_get(void)
{
  if (! context_stack)
    return NULL;

  return context_stack->data;
}


void                    
libnd_phdl_ht_add_child(LND_PHDL_HeaderTempl *header,
			LND_PHDL_FieldTempl *field)
{
  if (!header || !field)
    return;

  libnd_phdl_ft_add_child(header->root, field);
}


GList *
libnd_phdl_ht_get_children(LND_PHDL_HeaderTempl *header)
{
  LND_PHDL_Block *block;

  if (! header)
    {
      D(("Input error.\n"));
      return NULL;
    }

  block = header->root->content;
  return block->fields;
}


LND_Protocol           *
libnd_phdl_ht_create_proto(LND_PHDL_HeaderTempl *header)
{
  /* FIXME */
  return NULL;
  TOUCH(header);
}


void                    
libnd_phdl_ht_free(LND_PHDL_HeaderTempl *header)
{
  if (!header)
    return;

  /* Fields in context stack and demux field should be somewhere
   * in the tree contained in the "fields" list of the root block
   * template, so they will get cleaned up when that dummy field
   * template gets released, and not here.
   */

  libnd_phdl_ft_free(header->root);
  g_free(header);
}


static LND_PHDL_FieldTempl *
phdl_field_new(LND_PHDL_FieldType type,
	       const char *name)
{
  LND_PHDL_FieldTempl *field;

  if (! (field = g_new0(LND_PHDL_FieldTempl, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  field->type = type;
  field->name = g_strdup(name);

  return field;
}

static void
phdl_field_free(LND_PHDL_FieldTempl *field)
{
  if (!field)
    return;

  g_free(field->name);
  g_free(field);
}


LND_PHDL_FieldTempl *
libnd_phdl_ft_clone(const LND_PHDL_FieldTempl *ft, const char *name)
{
  LND_PHDL_FieldTempl *clone;
  LND_PHDL_FTCommon   *common;

  if (!ft)
    return NULL;

  if (! (clone = phdl_field_new(ft->type, name)))
    return NULL;

  clone->content = ft->content;

  common = (LND_PHDL_FTCommon *) ft->content;
  common->use_count++;

  return clone;  
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_block_new(const char *name)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_Block *block;

  if (! (field = phdl_field_new(LND_PHDL_BLOCK, name)))
    return NULL;

  if (! (block = g_new0(LND_PHDL_Block, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }

  block->common.use_count = 1;
  field->content = block;

  return field;
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_struct_new(const char *name)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_Struct *st;

  if (! (field = phdl_field_new(LND_PHDL_STRUCT, name)))
    return NULL;

  if (! (st = g_new0(LND_PHDL_Struct, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }

  field->content = st;

  return field;
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_chain_new(const char *name)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_Chain *chain;

  if (! (field = phdl_field_new(LND_PHDL_CHAIN, name)))
    return NULL;

  if (! (chain = g_new0(LND_PHDL_Chain, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }

  chain->common.use_count = 1;
  field->content = chain;

  return field;
}


void
libnd_phdl_ft_chain_set_end(LND_PHDL_FieldTempl *ft,
			 const LND_PHDL_ChainEnd *end)
{
  LND_PHDL_Chain *chain;

  if (!ft || !end)
    return;

  if (!ft->type == LND_PHDL_CHAIN)
    return;

  chain = ft->content;

  if (chain->end)
    libnd_phdl_ft_chainend_free((LND_PHDL_ChainEnd *) chain->end);
  
  chain->end = end;
}


LND_PHDL_ChainEnd      *
libnd_phdl_ft_chainend_new_items(LND_PHDL_NumExp *exp)
{
  LND_PHDL_ChainEnd *end;

  if (!exp)
    return NULL;

  if (! (end = g_new0(LND_PHDL_ChainEnd, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  end->type = LND_PHDL_CHAINEND_ITEMS;
  end->exp  = exp;

  D(("Created new chain end condition ITEMS: %s\n",
     libnd_phdl_numexp_to_string(exp)));

  return end;
}


LND_PHDL_ChainEnd      *
libnd_phdl_ft_chainend_new_length(LND_PHDL_NumExp *exp)
{
  LND_PHDL_ChainEnd *end;

  if (!exp)
    return NULL;

  if (! (end = g_new0(LND_PHDL_ChainEnd, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  end->type = LND_PHDL_CHAINEND_BITS;
  end->exp  = exp;

  D(("Created new chain end condition BITS: %s\n",
     libnd_phdl_numexp_to_string(exp)));

  return end;
}


LND_PHDL_ChainEnd      *
libnd_phdl_ft_chainend_new_pattern(LND_PHDL_FieldTempl *templ)
{
  LND_PHDL_ChainEnd *end;

  if (!templ)
    return NULL;
  
  if (! (end = g_new0(LND_PHDL_ChainEnd, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  /* FIXME */
  return end;
}


void                    
libnd_phdl_ft_chainend_free(LND_PHDL_ChainEnd *end)
{
  if (!end)
    return;

  switch (end->type)
    {
    case LND_PHDL_CHAINEND_ITEMS:
      libnd_phdl_numexp_free((LND_PHDL_NumExp *) end->exp);
      break;

    case LND_PHDL_CHAINEND_BITS:
      libnd_phdl_numexp_free((LND_PHDL_NumExp *) end->exp);
      break;

    case LND_PHDL_CHAINEND_PATTERN:
      {
	GList *l;

	/* Iterate over all assignment patterns and free them. */
	for (l = end->fieldexps; l; l = g_list_next(l))
	  libnd_phdl_boolexp_free(l->data);

	g_list_free(end->fieldexps);
	libnd_phdl_field_free(end->field);
      }
      break;

    default:
      D(("ERROR: invalid input.\n"));
    }

  g_free(end);
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_union_new(const char *name)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_Union *un;

  if (! (field = phdl_field_new(LND_PHDL_UNION, name)))
    return NULL;
  
  if (! (un = g_new0(LND_PHDL_Union, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }
  
  un->common.use_count = 1;
  field->content = un;

  return field;
}


void                    
libnd_phdl_ft_union_add_field(LND_PHDL_FieldTempl *union_field,
			   LND_PHDL_FieldTempl *field,
			   LND_PHDL_BoolExp *exp)
{
  LND_PHDL_Union *un = NULL;

  if (!union_field || !field || !exp)
    return;

  if (union_field->type != LND_PHDL_UNION)
    {
      D(("Invalid input -- not union.\n"));
      return;
    }

  un = union_field->content;
  un->fields = g_list_append(un->fields, field);
  un->bool_exps = g_list_append(un->bool_exps, exp);
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_enum_new(const char *name,
		    const LND_PHDL_NumExp *exp)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_Enum *en;

  if (! (field = phdl_field_new(LND_PHDL_ENUM, name)))
    return NULL;

  if (! (en = g_new0(LND_PHDL_Enum, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }
 
  field->content = en;
  en->common.use_count = 1;
  en->length = exp;
 
  return field;
}


void                    
libnd_phdl_ft_enum_set_hex_format(LND_PHDL_FieldTempl *field,
			       gboolean is_hex)
{
  LND_PHDL_Enum *en;

  if (!field)
    return;

  en = field->content;
  en->is_hex = is_hex;
}


void                    
libnd_phdl_ft_enum_add_item(LND_PHDL_FieldTempl *field,
			 LND_PHDL_EnumItem *item)
{
  LND_PHDL_Enum *en;

  if (!field || !item)
    return;

  en = field->content;
  en->items = g_list_append(en->items, item);
}


LND_PHDL_EnumItem      *
libnd_phdl_ft_enumitem_new(const char *label, gint64 value)
{
  LND_PHDL_EnumItem *item;

  if (! (item = g_new0(LND_PHDL_EnumItem, 1)))
    {
      D(("Out of memory\n"));
      return NULL;
    }

  item->label = g_strdup(label);
  item->value = value;

  return item;
}


void
libnd_phdl_ft_enumitem_free(LND_PHDL_EnumItem *item)
{
  if (!item)
    return;

  g_free(item->label);
  g_free(item);
}


LND_PHDL_FieldParams   *
libnd_phdl_ft_fieldparams_new(void)
{
  LND_PHDL_FieldParams *params = g_new0(LND_PHDL_FieldParams, 1);

  if (!params)
    {
      D(("Out of memory\n"));
      return NULL;
    }
    
  return params;
}

LND_PHDL_FieldParams        *
libnd_phdl_ft_fieldparams_copy(LND_PHDL_FieldParams *orig)
{
  LND_PHDL_FieldParams *copy;

  if (! (copy = libnd_phdl_ft_fieldparams_new()))
    return NULL;

  *copy = *orig;
  
  if (orig->sep)
    copy->sep = g_strdup(orig->sep);

  return copy;
}

void                    
libnd_phdl_ft_fieldparams_free(LND_PHDL_FieldParams *params)
{
  if (!params)
    return;

  g_free(params->sep);
  g_free(params);
}



LND_PHDL_FieldTempl    *
libnd_phdl_ft_int_new(const char *name, const LND_PHDL_NumExp *exp)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_NumField *num;

  if (! (field = phdl_field_new(LND_PHDL_INT, name)))
    return NULL;

  if (! (num = g_new0(LND_PHDL_NumField, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }

  field->content = num;
  num->common.use_count = 1;
  num->length = exp;

  return field;
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_hex_new(const char *name, const LND_PHDL_NumExp *exp)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_NumField *num;

  if (! (field = phdl_field_new(LND_PHDL_HEX, name)))
    return NULL;

  if (! (num = g_new0(LND_PHDL_NumField, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }

  field->content = num;
  num->common.use_count = 1;
  num->length = exp;

  return field;
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_str_fixed_new(const char *name, const LND_PHDL_NumExp *exp)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_StrField *str;

  if (! (field = phdl_field_new(LND_PHDL_STR, name)))
    return NULL;

  if (! (str = g_new0(LND_PHDL_StrField, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }

  field->content = str;
  str->common.use_count = 1;
  str->length = exp;

  return field;
}


LND_PHDL_FieldTempl    *
libnd_phdl_ft_str_variable_new(const char *name, const char *end_marker)
{
  LND_PHDL_FieldTempl *field;
  LND_PHDL_StrField *str;

  if (! (field = phdl_field_new(LND_PHDL_STR, name)))
    return NULL;

  if (! (str = g_new0(LND_PHDL_StrField, 1)))
    {
      phdl_field_free(field);
      return NULL;
    }

  field->content = str;
  str->common.use_count = 1;
  str->end_marker = g_strdup(end_marker);

  return field;
}


gboolean                
libnd_phdl_ft_set_params(LND_PHDL_FieldTempl *templ,
			 LND_PHDL_FieldParams *params)
{
  if (!templ || !params)
    return FALSE;

  switch (templ->type)
    {
    case LND_PHDL_INT:
    case LND_PHDL_HEX:
      {
	LND_PHDL_NumField *num = templ->content;
	
	if (num->params)
	  libnd_phdl_ft_fieldparams_free(num->params);

	num->params = libnd_phdl_ft_fieldparams_copy(params); 
      }
      break;
      
    case LND_PHDL_STR:
      {
	LND_PHDL_StrField *str = templ->content;
	
	if (str->params)
	  libnd_phdl_ft_fieldparams_free(str->params);

	str->params = libnd_phdl_ft_fieldparams_copy(params); 
      }
      break;

    case LND_PHDL_BLOCK:
    case LND_PHDL_CHAIN:
    case LND_PHDL_UNION:
    case LND_PHDL_ENUM:
    default:
      D(("Cannot set parameters for this field type.\n"));
      return FALSE;
    }

  return TRUE;
}


const LND_PHDL_FieldParams   *
libnd_phdl_ft_get_params(LND_PHDL_FieldTempl *templ)
{
  if (!templ)
    return NULL;

  switch (templ->type)
    {
    case LND_PHDL_INT:
    case LND_PHDL_HEX:
      {
	LND_PHDL_NumField *num = templ->content;
	return num->params;
      }
      break;
      
    case LND_PHDL_STR:
      {
	LND_PHDL_StrField *str = templ->content;
	return str->params;
      }
      break;

    case LND_PHDL_BLOCK:
    case LND_PHDL_CHAIN:
    case LND_PHDL_UNION:
    case LND_PHDL_ENUM:
    default:
      D(("Cannot set parameters for this field type.\n"));
    }  

  return NULL;
}


void                    
libnd_phdl_ft_add_child(LND_PHDL_FieldTempl *parent,
			LND_PHDL_FieldTempl *child)
{
  if (!parent || !child)
    return;

  switch (parent->type)
    {
    case LND_PHDL_BLOCK:
      {
	LND_PHDL_Block *block = parent->content;
	block->fields = g_list_append(block->fields, child);
      }
      break;

    case LND_PHDL_CHAIN:
      {
	LND_PHDL_Block *chain = parent->content;
	chain->fields = g_list_append(chain->fields, child);
      }
      break;

    case LND_PHDL_UNION:
    case LND_PHDL_ENUM:
    case LND_PHDL_INT:
    case LND_PHDL_HEX:
    case LND_PHDL_STR:
    default:
      D(("Cannot add child to field type %u\n", parent->type));
    }  
}

void
libnd_phdl_ft_free(LND_PHDL_FieldTempl *templ)
{
  LND_PHDL_FTCommon *com;

  if (!templ)
    return;

  com = (LND_PHDL_FTCommon *) templ->content;
  
  if (--com->use_count == 0)
    {
      switch (templ->type)
	{
	case LND_PHDL_INT:
	case LND_PHDL_HEX:
	  {
	    LND_PHDL_NumField *num = templ->content;
	    
	    libnd_phdl_numexp_free((LND_PHDL_NumExp *) num->length);
	    libnd_phdl_ft_fieldparams_free(num->params);
	    g_free(num);
	  }
	  break;
	  
	case LND_PHDL_STR:
	  {
	    LND_PHDL_StrField *str = templ->content;
	    
	    libnd_phdl_numexp_free((LND_PHDL_NumExp *) str->length);
	    libnd_phdl_ft_fieldparams_free(str->params);
	    g_free(str->end_marker);
	    g_free(str);
	  }
	  break;
	  
	case LND_PHDL_BLOCK:
	  {
	    LND_PHDL_Block *block = templ->content;	    
	    GList *l;
	    
	    for (l = block->fields; l; l = g_list_next(l))
	      libnd_phdl_ft_free(l->data);
	    
	    g_list_free(block->fields);
	    g_free(block);	
	  }
	  break;
	  
	case LND_PHDL_CHAIN:
	  {
	    LND_PHDL_Chain *chain = templ->content;	    
	    GList *l;
	    
	    for (l = chain->fields; l; l = g_list_next(l))
	      libnd_phdl_field_free(l->data);
	    
	    libnd_phdl_ft_chainend_free((LND_PHDL_ChainEnd *) chain->end);
	    g_list_free(chain->fields);
	    g_free(chain);
	  }
	  break;
	  
	case LND_PHDL_UNION:
	  {
	    LND_PHDL_Union *un = templ->content;	    
	    GList *l;
		
	    for (l = un->fields; l; l = g_list_next(l))
	      libnd_phdl_ft_free(l->data);
	    for (l = un->bool_exps; l; l = g_list_next(l))
	      libnd_phdl_boolexp_free(l->data);
	    
	    g_list_free(un->fields);
	    g_list_free(un->bool_exps);
	    g_free(un);
	  }
	  break;
	  
	case LND_PHDL_ENUM:
	  {
	    LND_PHDL_Enum *en = templ->content;	    
	    GList *l;
	    
	    for (l = en->items; l; l = g_list_next(l))
	      libnd_phdl_ft_enumitem_free(l->data);
	    
	    libnd_phdl_numexp_free((LND_PHDL_NumExp*) en->length);
	    g_list_free(en->items);
	    g_free(en);
	  }
	  break;
	  
	default:
	  D(("Cannot set parameters for this field type.\n"));
	  return;
	}
      
      templ->content = NULL;
    }
  
  g_free(templ->name);
  g_free(templ);
}


LND_Protocol           *
libnd_phdl_get_header_from_proto(LND_PHDL_HeaderTempl *header)
{
  /* FIMXE */
  return NULL;
  TOUCH(header);
}


LND_PHDL_HeaderTempl   *
libnd_phdl_get_proto_from_header(LND_Protocol *proto)
{
  /* FIXME */
  return NULL;
  TOUCH(proto);
}


void                    
libnd_phdl_ht_ctxt_set(LND_PHDL_HeaderTempl *ht)
{
  if (ht)
    context_header = ht;
}


LND_PHDL_HeaderTempl   *
libnd_phdl_ht_ctxt_get(void)
{
  return context_header;
}
