/*

Copyright (C) 2004 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "libnd_phdl.h"
#include "libnd_phdl_header.h"


LND_PHDL_NumExp    *
libnd_phdl_numexp_new(LND_PHDL_NumParam *param1,
		      LND_PHDL_NumParam *param2,
		      LND_PHDL_NumOp op)
{
  LND_PHDL_NumExp *exp;

  if (!param1)
    {
      D(("Invalid input\n"));
      return NULL;
    }

  if (! (exp = g_new0(LND_PHDL_NumExp, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }
  
  exp->param1 = param1;
  exp->param2 = param2;
  exp->op = op;

  return exp;
}


void                
libnd_phdl_numexp_free(LND_PHDL_NumExp *exp)
{
  if (!exp)
    return;

  libnd_phdl_numparam_free(exp->param1);
  libnd_phdl_numparam_free(exp->param2);
  
  g_free(exp);
}


LND_PHDL_BoolExp   *
libnd_phdl_boolexp_new(LND_PHDL_NumParam *param1,
		       LND_PHDL_NumParam *param2,
		       LND_PHDL_BoolOp op)
{
  LND_PHDL_BoolExp *exp;

  if (!param1)
    {
      D(("Invalid input\n"));
      return NULL;
    }

  if (! (exp = g_new0(LND_PHDL_BoolExp, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }
  
  exp->param1 = param1;
  exp->param2 = param2;
  exp->op = op;

  return exp;
}


void                
libnd_phdl_boolexp_free(LND_PHDL_NumExp *exp)
{
  if (!exp)
    return;

  libnd_phdl_numparam_free(exp->param1);
  libnd_phdl_numparam_free(exp->param2);
  
  g_free(exp);
}


LND_PHDL_NumParam  *
libnd_phdl_numparam_lit_new(int value)
{
  LND_PHDL_NumParam *param;

  if (! (param = g_new0(LND_PHDL_NumParam, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }
  
  param->type = LND_PHDL_LITERAL;
  param->literal = value;

  return param;
}


LND_PHDL_NumParam  *
libnd_phdl_numparam_exp_new(LND_PHDL_NumExp *exp)
{
  LND_PHDL_NumParam *param;

  if (! (param = g_new0(LND_PHDL_NumParam, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }
  
  param->type = LND_PHDL_EXP;
  param->exp = exp;

  return param;
}


LND_PHDL_NumParam  *
libnd_phdl_numparam_field_new(const char *field_name)
{
  LND_PHDL_NumParam *param;

  if (! (param = g_new0(LND_PHDL_NumParam, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }
  
  param->type = LND_PHDL_FIELD;
  param->field_name = g_strdup(field_name);

  return param;
}


void                
libnd_phdl_numparam_free(LND_PHDL_NumParam *param)
{
  if (!param)
    return;

  switch (param->type)
    {
    case LND_PHDL_LITERAL:
      break;

    case LND_PHDL_FIELD:
      g_free(param->field_name);
      break;

    case LND_PHDL_EXP:
      libnd_phdl_numexp_free(param->exp);
      break;

    default:
      D(("Error -- invalid case\n"));
    }

  g_free(param);
}


gboolean            
libnd_phdl_numparam_eval(LND_PHDL_NumParam *param,
			 LND_PHDL_Header *header,
			 guint64 *result)
{
  if (!param || !result)
    return FALSE;

  switch (param->type)
    {
    case LND_PHDL_LITERAL:
      *result = param->literal;
      break;

    case LND_PHDL_FIELD:
      {
	LND_PHDL_Field *field;

	if (! (field = libnd_phdl_header_find_field(header, param->field_name)))
	  {
	    D(("Field %s not found.\n", param->field_name));
	    return FALSE;
	  }
	
	if (! libnd_phdl_field_eval(field, result))
	  {
	    D(("Field evaluation of %s failed.\n", param->field_name));
	    return FALSE;
	  }
      }
      break;
      
    case LND_PHDL_EXP:
      if (! libnd_phdl_numexp_eval(param->exp, header, result))
	{
	  D(("Evalutaion of subexpression failed.\n"));
	  return FALSE;
	}
      break;

    default:
      D(("Error -- invalid case %i\n", param->type));
    }
  
  return TRUE;
}


const char         *
libnd_phdl_numparam_to_string(const LND_PHDL_NumParam *param)
{
  char str[MAXPATHLEN];
  const char *result = str;
  
  if (! param)
    return "(null)";

  switch (param->type)
    {
    case LND_PHDL_LITERAL:
      g_snprintf(str, MAXPATHLEN, "%i", param->literal);
      break;
      
    case LND_PHDL_FIELD:
      result = param->field_name;
      break;
      
    case LND_PHDL_EXP:
    default:
      result = libnd_phdl_numexp_to_string(param->exp);
    }

  return result;
}


gboolean         
libnd_phdl_numexp_eval(const LND_PHDL_NumExp *exp,
		       LND_PHDL_Header *header,
		       guint64 *result)
{
  guint64 result1, result2;
  gboolean success1, success2;
  
  if (!exp || !result)
    return FALSE;

  success1 = libnd_phdl_numparam_eval(exp->param1, header, &result1);

  if (! exp->param2)
    {
      *result = result1;
      return success1;
    }

  success2 = libnd_phdl_numparam_eval(exp->param2, header, &result2);
  
  if (!success1 || !success2)
    return FALSE;

  switch (exp->op)
    {
    case LND_PHDL_ADD:
      *result = result1 + result2;
      break;

    case LND_PHDL_SUB:
      *result = result1 - result2;
      break;

    case LND_PHDL_MUL:
      *result = result1 * result2;
      break;

    case LND_PHDL_DIV:
      *result = result1 / result2;
      break;

    case LND_PHDL_MOD:
      *result = result1 % result2;
      break;

    default:
      D(("Invalid operator value.\n"));
    }

  return TRUE;
}


const char         *
libnd_phdl_numexp_to_string(const LND_PHDL_NumExp *exp)
{
  char str1[MAXPATHLEN];
  char str2[MAXPATHLEN];
  static char str3[MAXPATHLEN];

  if (!exp)
    return "(null)";

  if (exp->op == LND_PHDL_NON)
    return libnd_phdl_numparam_to_string(exp->param1);

  g_snprintf(str1, MAXPATHLEN, "%s", libnd_phdl_numparam_to_string(exp->param1));
  g_snprintf(str2, MAXPATHLEN, "%s", libnd_phdl_numparam_to_string(exp->param2));
  
  switch (exp->op)
    {
    case LND_PHDL_ADD:
      g_snprintf(str3, MAXPATHLEN, "(%s + %s)", str1, str2);
      break;
      
    case LND_PHDL_SUB:
      g_snprintf(str3, MAXPATHLEN, "(%s - %s)", str1, str2);
      break;
      
    case LND_PHDL_MUL:
      g_snprintf(str3, MAXPATHLEN, "(%s * %s)", str1, str2);
      break;
      
    case LND_PHDL_DIV:
      g_snprintf(str3, MAXPATHLEN, "(%s / %s)", str1, str2);
      break;
      
    case LND_PHDL_MOD:
      g_snprintf(str3, MAXPATHLEN, "(%s %% %s)", str1, str2);
      break;
      
    default:
      D(("Input error -- invalid operator %i\n", exp->op));
      g_snprintf(str3, MAXPATHLEN, "(%s <err> %s)", str1, str2);
    }

  return str3;
}


gboolean            
libnd_phdl_boolexp_eval(const LND_PHDL_BoolExp *exp,
			LND_PHDL_Header *header,
			gboolean *result)
{
  guint64 result1, result2;
  gboolean success1, success2;
  
  if (!exp || !result)
    return FALSE;

  success1 = libnd_phdl_numparam_eval(exp->param1, header, &result1);

  if (! exp->param2)
    {
      *result = result1;
      return success1;
    }

  success2 = libnd_phdl_numparam_eval(exp->param2, header, &result2);
  
  if (!success1 || !success2)
    return FALSE;

  switch (exp->op)
    {
    case LND_PHDL_EQU:
      *result = (result1 == result2);
      break;

    case LND_PHDL_NEQ:
      *result = (result1 != result2);
      break;

    case LND_PHDL_LET:
      *result = (result1 <  result2);
      break;

    case LND_PHDL_GET:
      *result = (result1 >  result2);
      break;

    case LND_PHDL_LEQ:
      *result = (result1 <= result2);
      break;

    case LND_PHDL_GEQ:
      *result = (result1 >= result2);
      break;

    default:
      D(("Invalid operator value.\n"));
    }
  
  return TRUE;
}


const char         *
libnd_phdl_boolexp_to_string(const LND_PHDL_BoolExp *exp)
{
  char str1[MAXPATHLEN];
  char str2[MAXPATHLEN];
  static char str3[MAXPATHLEN];

  if (!exp)
    return "(null)";

  if (exp->op == LND_PHDL_NOB)
    return libnd_phdl_numparam_to_string(exp->param1);

  g_snprintf(str1, MAXPATHLEN, "%s", libnd_phdl_numparam_to_string(exp->param1));
  g_snprintf(str2, MAXPATHLEN, "%s", libnd_phdl_numparam_to_string(exp->param2));
  
  switch (exp->op)
    {
    case LND_PHDL_EQU:
      g_snprintf(str3, MAXPATHLEN, "(%s = %s)", str1, str2);
      break;
      
    case LND_PHDL_NEQ:
      g_snprintf(str3, MAXPATHLEN, "(%s != %s)", str1, str2);
      break;
      
    case LND_PHDL_LET:
      g_snprintf(str3, MAXPATHLEN, "(%s < %s)", str1, str2);
      break;
      
    case LND_PHDL_GET:
      g_snprintf(str3, MAXPATHLEN, "(%s > %s)", str1, str2);
      break;
      
    case LND_PHDL_LEQ:
      g_snprintf(str3, MAXPATHLEN, "(%s <= %s)", str1, str2);
      break;

    case LND_PHDL_GEQ:
      g_snprintf(str3, MAXPATHLEN, "(%s >= %s)", str1, str2);
      break;
      
    default:
      D(("Input error -- invalid operator %i\n", exp->op));
      g_snprintf(str3, MAXPATHLEN, "(%s <err> %s)", str1, str2);
    }

  return str3;
}

