/*

Copyright (C) 2004 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <libnd.h>

#include "libnd_phdl.h"
#include "libnd_phdl_header.h"


static const char *
phdl_header_name_get(const char *input, char *result, int len)
{
  int input_len;
  const char *dot, *next;
  
  input_len = strlen(input);
  next = input + input_len;

  if ( (dot = strchr(input, '.')))
    {
      input_len = dot - input;
      next = input + input_len + 1;
    }
  
  memcpy(result, input, MIN(len, input_len));  
  result[len] = '\0';
  
  return next;
}


static LND_PHDL_Field *
phdl_field_find_in_list(GList *l, const char *field_name)
{
  char subname[MAXPATHLEN];
  const char *next;

  next = phdl_header_name_get(field_name, subname, MAXPATHLEN);
  D(("Looking for component %s of %s\n", subname, field_name));
  
  for (/* l given in params */; l; l = g_list_next(l))
    {
      LND_PHDL_Field *field = l->data;
      
      if (! strcmp(subname, field->templ->name))
	{
	  D(("Component %s found.\n", subname));
	  
	  if (*next != '\0')
	    {
	      D(("Recursing to find remaining %s\n", next));
	      return phdl_field_find_in_list(field->fields, next);
	    }
	  
	  D(("End of string -- returning field named %s\n", field->templ->name));
	  return field;
	}
    }
  
  D(("Field %s not found.\n", field_name));
  return NULL;
}


static gboolean
phdl_build_instance_from_list(LND_PHDL_Header *header,
			      GList *field_templs)
{
  GList *l, *recurse_fields = NULL;
  LND_PHDL_Field *field;
  LND_PHDL_Field *ctxt;

  /* Get the topmost element of the context stack -- that's the
   * field that we need to fill with new fields.
   */
  ctxt = (LND_PHDL_Field *) header->context->data;

  /* Iterate over all the given field templates and add a new field
   * for each of these templates to the context field.
   */
  for (l = field_templs; l; l = g_list_next(l))
    {
      LND_PHDL_FieldTempl *field_templ = l->data;

      switch (field_templ->type)
	{
	case LND_PHDL_INT:
	case LND_PHDL_HEX:
	  if (! (field = libnd_phdl_field_new_num(field_templ, &header->bitptr)))
	    return FALSE;
	  break;
	  
	case LND_PHDL_STR:
	  if (! (field = libnd_phdl_field_new_str(field_templ, &header->bitptr)))
	    return FALSE;
	  break;
	  
	case LND_PHDL_BLOCK:
	  {
	    LND_PHDL_Block *block = field_templ->content;
	    
	    if (! (field = libnd_phdl_field_new_block(field_templ, &header->bitptr)))
	      return FALSE;

	    recurse_fields = block->fields;
	  }
	  break;

	case LND_PHDL_CHAIN:
	  {
	    LND_PHDL_Chain *chain = field_templ->content;

	    if (! (field = libnd_phdl_field_new_chain(field_templ, &header->bitptr)))
	      return FALSE;
	    
	    recurse_fields = chain->fields;
	  }
	  break;

	case LND_PHDL_UNION:
	  {
	    LND_PHDL_Union *un = field_templ->content;

	    if (! (field = libnd_phdl_field_new_union(field_templ, &header->bitptr)))
	      return FALSE;
	    
	    recurse_fields = un->fields;
	  }
	  break;

	case LND_PHDL_ENUM:
	  if (! (field = libnd_phdl_field_new_enum(field_templ, &header->bitptr)))
	    return FALSE;
	  break;
	  
	default:
	  D(("Invalid field type %u\n", field_templ->type));
	  return FALSE;
	}
      
      ctxt->fields = g_list_append(ctxt->fields, field);
      
      if (recurse_fields)
	{
	  header->context = g_list_prepend(header->context, field);
	  
	  if (! phdl_build_instance_from_list(header, recurse_fields))
	    {
	      D(("Error in recursive field parsing.\n"));
	      return FALSE;
	    }
	}
    }

  /* After we're done with that, remove the topmost element from
   * the context stack.
   */
  header->context = g_list_remove_link(header->context, header->context);
  
  return TRUE;
}



LND_PHDL_Header *
libnd_phdl_header_new(LND_PHDL_HeaderTempl *templ,
		      LND_ProtoData *data)
{
  LND_PHDL_Header *header;

  if (!templ || !data)
    return NULL;

  if (! (header = g_new0(LND_PHDL_Header, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }
  
  header->templ = templ;

  header->bitptr.header   = header;
  header->bitptr.data     = data->data;
  header->bitptr.data_end = data->data_end;
  header->bitptr.off      = 0;

  if (! (header->root = g_new0(LND_PHDL_Field, 1)))
    {
      D(("Out of memory.\n"));
      g_free(header);
      return NULL;
    }

  header->context = g_list_prepend(header->context, header->root);
  
  /* This is where the fun is -- we now initialize the header instance
   * using @data, as specified in the template passed in @templ.
   */
  phdl_build_instance_from_list(header, libnd_phdl_ht_get_children(templ));
  
  return header;
}


void                    
libnd_phdl_header_free(LND_PHDL_Header *header)
{
  GList *l;

  if (!header)
    return;

  for (l = header->root->fields; l; l = g_list_next(l))
    libnd_phdl_field_free((LND_PHDL_Field *) l->data);

  g_free(header->root);
  g_free(header);
}


LND_PHDL_Field *
libnd_phdl_header_find_field(LND_PHDL_Header *header,
			     const char *field_name)
{
  if (!header || !field_name || !*field_name)
    return NULL;

  return phdl_field_find_in_list(header->root->fields, field_name);
}


gboolean
libnd_phdl_header_get_field_length(LND_PHDL_Header *header,
				   LND_PHDL_FieldTempl *templ,
				   guint *result_ptr)
{
  guint64 result = 0;

  if (!header || !templ || !result_ptr)
    {
      D(("Input error.\n"));
      return 0;
    }

  /* We're given a partially initialized header with field values
   * that we can use, and a field template that tells us how we can
   * derive a field's size.
   */
  switch (templ->type)
    {
    case LND_PHDL_INT:
    case LND_PHDL_HEX:
      {
	LND_PHDL_NumField *num = templ->content;

	if (! (libnd_phdl_numexp_eval(num->length, header, &result)))
	  {
	    D(("Numerical field evaluation error.\n"));
	    return FALSE;
	  }
      }
      break;
      
    case LND_PHDL_STR:
      {
	LND_PHDL_StrField *str = templ->content;
	
	if (str->end_marker)
	  {
	    D(("Dynamic-length string field.\n"));
	    return FALSE;
	  }
	else
	  {
	    if (! (libnd_phdl_numexp_eval(str->length, header, &result)))
	      {
		D(("String field evaluation error.\n"));
		return FALSE;
	      }
	  }
      }
      break;

    case LND_PHDL_BLOCK:
      {
	LND_PHDL_Block *block = templ->content;
	guint tmp;
	GList *l;
	
	/* Iterate over all the fields contained in the template
	 * and sum up their lengths.
	 */
	for (l = block->fields; l; l = g_list_next(l))
	  {
	    if (! (libnd_phdl_header_get_field_length(header, l->data, &tmp)))
	      {
		D(("Error in calculating field length.\n"));
		return FALSE;
	      }

	    result += tmp;
	  }
      }
      break;

    case LND_PHDL_CHAIN:
      {
	LND_PHDL_Chain *chain = templ->content;
	
	/* For a chain, we want to know the entire space we
	 * have available -- looking at the actual children
	 * is not necessary:
	 */
	if (chain->end->type == LND_PHDL_CHAINEND_ITEMS ||
	    chain->end->type == LND_PHDL_CHAINEND_PATTERN)
	  {
	    D(("Dynamically sized chain.\n"));
	    return FALSE;
	  }

	if (! (libnd_phdl_numexp_eval(chain->end->exp, header, &result)))
	  {
	    D(("Chain length evaluation error.\n"));
	    return FALSE;
	  }
      }
      break;
      
    case LND_PHDL_UNION:
      D(("Size of a union set of fields unknown until evaluation\n"));
      return FALSE;
      
    case LND_PHDL_ENUM:
      {
	LND_PHDL_Enum *en = templ->content;
	
	if (! (libnd_phdl_numexp_eval(en->length, header, &result)))
	  {
	    D(("Chain length evaluation error.\n"));
	    return FALSE;
	  }
      }
      break;
      
    default:
      D(("Illegal field type %i\n", templ->type));
      return FALSE;
    }
  
  *result_ptr = (guint) result;
  return TRUE;
}


LND_PHDL_Field         *
libnd_phdl_field_new_num(LND_PHDL_FieldTempl *templ, 
			 LND_PHDL_BitPtr *ptr)
{
  LND_PHDL_Field *field;

  if (! templ || templ->type != LND_PHDL_INT)
    {
      D(("Input error.\n"));
      return NULL;
    }

  if (! (field = g_new0(LND_PHDL_Field, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  field->templ = templ;

  if (! (libnd_phdl_header_get_field_length(ptr->header, templ,
					    &field->length)))
    {
      D(("Field length evaluation failed.\n"));
      libnd_phdl_field_free(field);
      return NULL;
    }
  
  if (! (libnd_phdl_bitptr_get_raw(ptr, field->length,
				   (guchar*) &field->val.int_val,
				   sizeof(field->val.int_val))))
    {
      D(("Data extraction error\n"));
      libnd_phdl_field_free(field);
      return NULL;
    }
  
  return field;
}


LND_PHDL_Field         *
libnd_phdl_field_new_str(LND_PHDL_FieldTempl *templ,
			 LND_PHDL_BitPtr *ptr)
{
  LND_PHDL_Field *field;
  LND_PHDL_StrField *str;
  char result[MAXPATHLEN];
  
  if (! templ || templ->type != LND_PHDL_STR)
    {
      D(("Input error.\n"));
      return NULL;
    }

  if (! (field = g_new0(LND_PHDL_Field, 1)))
    {
      D(("Out of memory.\n"));
      return NULL;
    }

  str = templ->content;
  field->templ = templ;

  /* If the template has an end-of-string marker, the length is
   * dynamic and we need to hunt for the marker. Otherwise, the
   * length is fixed and we obtain it by evaluating the length
   * expression.
   */
  if (str->end_marker)
    {
      if (! libnd_phdl_bitptr_get_string_marked(ptr, str->end_marker,
						result, MAXPATHLEN))
	{
	  D(("String extraction failed.\n"));
	  g_free(field);
	  return NULL;
	}
    }
  else
    {
      if (! (libnd_phdl_header_get_field_length(ptr->header, templ,
						&field->length)))
	{
	  D(("Field length evaluation failed.\n"));
	  libnd_phdl_field_free(field);
	  return NULL;
	}
      
      if (! libnd_phdl_bitptr_get_string_fixed(ptr,
					       MAX(field->length, MAXPATHLEN),
					       result))
	{
	  D(("String extraction failed.\n"));
	  g_free(field);
	  return NULL;
	}
    }
  
  field->val.str_val = g_strdup(result);
  
  return field;
}


LND_PHDL_Field         *
libnd_phdl_field_new_block(LND_PHDL_FieldTempl *templ,
			   LND_PHDL_BitPtr *ptr)
{
  return NULL;
  TOUCH(templ); TOUCH(ptr);
}


LND_PHDL_Field         *
libnd_phdl_field_new_chain(LND_PHDL_FieldTempl *templ,
			   LND_PHDL_BitPtr *ptr)
{
  return NULL;
  TOUCH(templ); TOUCH(ptr);
}


LND_PHDL_Field         *
libnd_phdl_field_new_union(LND_PHDL_FieldTempl *templ,
			   LND_PHDL_BitPtr *ptr)
{
  return NULL;
  TOUCH(templ); TOUCH(ptr);
}

LND_PHDL_Field         *
libnd_phdl_field_new_enum(LND_PHDL_FieldTempl *templ,
			  LND_PHDL_BitPtr *ptr)
{
  return NULL;
  TOUCH(templ); TOUCH(ptr);
}


void                    
libnd_phdl_field_free(LND_PHDL_Field *field)
{
  GList *l;

  if (!field)
    return;

  if (field->templ)
    {
      switch (field->templ->type)
	{
	case LND_PHDL_STR:
	  g_free(field->val.str_val);
	  break;

	default:
	  /* Nothing to do */
	}
    }
  
  for (l = field->fields; l; l = g_list_next(l))
    libnd_phdl_field_free((LND_PHDL_Field *) l->data);
  
  g_list_free(field->fields);
  g_free(field);
}


LND_PHDL_Field *
libnd_phdl_field_find_field(LND_PHDL_Field *field,
			    const char *field_name)
{
  if (!field || !field_name || !*field_name)
    return NULL;

  return phdl_field_find_in_list(field->fields, field_name);
}


gboolean
libnd_phdl_field_eval(LND_PHDL_Field *field, guint64 *result)
{
  /* FIXME */
  return TRUE;
  TOUCH(field); TOUCH(result);
}


gboolean
libnd_phdl_bitptr_get_raw(LND_PHDL_BitPtr *ptr, guint bits,
			  guchar *dest, guint destbytes)
{
  guint byte_off, bit_off;
  guchar *src;
  int i;

  if (!ptr || !dest || (destbytes * 8) < bits)
    {
      D(("Input error.\n"));
      return FALSE;
    }

  /* Zero out destination first */
  memset(dest, 0, sizeof(guchar) * destbytes);

  byte_off = ptr->off / 8;
  bit_off  = ptr->off % 8;
  
  /* Sanity-check byte region -- make sure we don't go beyond what
   * we're supposed to look at:
   */
  if (ptr->data + byte_off > ptr->data_end)
    {
      D(("End of region reached in bitpointer.\n"));
      return FALSE;
    }

  src = ptr->data + byte_off;

  /* Let's optimise if we're on a byte boundary -- will mostly
   * be the case anyway! The only exception I see is when bitfields
   * are used. In that case we do bitwise iteration. Slow, but
   * since we're talking only a single bit or at most a small handful,
   * that should be okay.
   */
  if (bit_off == 0)
    {
      D(("Byte-aligned copy of %u bits.\n", bits));
      
      while (bits > 7 && src < ptr->data_end)
	{
	  *dest = *src;
	  bits -= 8;
	  
	  dest++; src++;
	}

      /* We end on an odd boundary -- handle that manually */
      while (bits > 0)
	{
	  *dest |= (*src & (1 << (bits - 1)));

	  bits--;
	}

      return TRUE;
    }

  /* Adjust bit offset so that we index the bit, not the
   * diversion from 8-bit boundaries. Example:
   *
   * 9 % 8 = 1, that'd be the first bit in byte 2,
   * that first bit has index 0.
   */
  bit_off--;
  D(("Unaligned copy of %u bits.\n", bits));

  /* Loop while there are still bits to copy. */
  while (bits > 0)
    {
      /* This one loops in byte-size chunks for the destination bytes.
       * Stop when we've copied enough or we run out of space.
       * i is the bit we're currently copying, bits is just the number
       * of bits we still have to do, and bit_off is the bit in the
       * source we're currently copying (since the bit index in the
       * source and destination aren't equal.
       */
      for (i = 0; bits > 0 && dest < ptr->data_end; i++, bits--, bit_off++)
	{
	  if (bit_off == 8)
	    {
	      bit_off = 0;
	      src++;
	    }
	  
	  *dest |= ((*src & (1 << bit_off)) << i);
	}
      
      dest++; bit_off = 0;
    }
  
  return TRUE;
}


static guchar *
phdl_bitptr_get_data(LND_PHDL_BitPtr *ptr)
{
  guint byte_off, bit_off;

  if (!ptr)
    return NULL;

  byte_off = ptr->off / 8;
  bit_off  = ptr->off % 8;

  if (byte_off >= (guint) (ptr->data_end - ptr->data))
    {
      D(("Invalid pointer offset\n"));
      return NULL;
    }

  if (bit_off != 0)
    {
      D(("Strings starting at non-byte boundaries not supported yet.\n"));
      return NULL;
    }

  return ptr->data + byte_off;
}


gboolean
libnd_phdl_bitptr_get_string_fixed(LND_PHDL_BitPtr *ptr,
				   guint length,
				   char *dest)
{
  guchar *src;

  if (!ptr || !dest)
    {
      D(("Input error.\n"));
      return FALSE;
    }

  if ( ! (src = phdl_bitptr_get_data(ptr)))
    return FALSE;

  /* Make sure that we don't copy more than we can provide */
  if (src + length >= ptr->data_end)
    {
      D(("%u bytes requested, can only provide %u.\n", length,
	 ptr->data_end - (src + length)));
      length = ptr->data_end - (src + length);
    }

  memcpy(dest, ptr->data, sizeof(guchar) * length);
  dest[length] = '\0';
  
  return TRUE;
}


gboolean
libnd_phdl_bitptr_get_string_marked(LND_PHDL_BitPtr *ptr,
				    const char *marker,
				    char *dest, int destlen)
{
  char *src;
  guint markerlen;

  if (!ptr || !marker || !dest)
    {
      D(("Input error.\n"));
      return FALSE;
    }

  if ( ! (src = phdl_bitptr_get_data(ptr)))
    return FALSE;
  
  markerlen = strlen(marker);

  while ((guchar *) src < (ptr->data_end - markerlen) && destlen > 0)
    {
      if (strstr(src, marker) == src)
	{
	  *dest = '\0';
	  return TRUE;
	}

      *dest++ = *src++;
      destlen--;
    }

  return FALSE;
}
