/*

Copyright (C) 2004 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifndef __libnd_phdl_header_h
#define __libnd_phdl_header_h

#include <libnd_phdl.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Instances of LND_PHDL_Field are actual fields contained
 * in a protocol header, in contrast to abstract descriptions.
 * LND_PHDL_Field instances contain actual header field values,
 * obtained when parsing a protocol header.
 */
struct lnd_phdl_field
{
  /* The field template that describes what this field
   * is an instance of.
   */
  LND_PHDL_FieldTempl  *templ;

  /* The total length of this particular field instance,
   * in bits.
   */
  guint                 length;
  
  /* Contained subfields, if any. */
  GList                *fields;

  /* The value of this field instance, unless this is
   * just a container for other subfields.
   */
  union {
    gint64              int_val;
    guchar             *str_val; /* length determinded through @length */
  } val;
};


/* LND_PHDL_BitPtrs are bit-granularity offset pointers
 * into a bytestring.
 */
struct lnd_phdl_bit_ptr
{
  LND_PHDL_Header      *header;

  guchar               *data;
  guchar               *data_end;

  int                   off;
};


/* LND_PHDL_Headers are instances of interpreted protocol
 * data and the entrance point for accessing these parsed
 * header fields. LND_PHDL_Headers use LND_PHDL_HeaderTempls
 * as templates for header layout -- these templates exist
 * once, whereas LND_PHDL_Headers exist once per interpreted
 * packet data.
 */
struct lnd_phdl_header
{
  /* Template defining header layout for this protocol.
   */
  LND_PHDL_HeaderTempl *templ;

  /* Pointer to data and bit offset into that data. Gets
   * updated on the fly while we're building the header
   * instance.
   */
  LND_PHDL_BitPtr       bitptr;

  /* The context list is used as a stack (operating on the front
   * of the list), with the first element being the field we're
   * currently analysing. This is needed
   * in case of nested fields.
   */
  GList                *context;

  /* The toplevel field of this header */
  LND_PHDL_Field       *root;

  /* Pointer to the demuxer field that guides further analysis
   * of protocol data.
   */
  LND_PHDL_Field       *demux_field;
};


LND_PHDL_Header        *libnd_phdl_header_new(LND_PHDL_HeaderTempl *templ,
					      LND_ProtoData *data);
void                    libnd_phdl_header_free(LND_PHDL_Header *header);

LND_PHDL_Field         *libnd_phdl_header_find_field(LND_PHDL_Header *header,
						     const char *field);

gboolean                libnd_phdl_header_get_field_length(LND_PHDL_Header *header,
							   LND_PHDL_FieldTempl *templ,
							   guint *result);
     
LND_PHDL_Field         *libnd_phdl_field_new_num(LND_PHDL_FieldTempl *templ, LND_PHDL_BitPtr *ptr);
LND_PHDL_Field         *libnd_phdl_field_new_str(LND_PHDL_FieldTempl *templ, LND_PHDL_BitPtr *ptr);
LND_PHDL_Field         *libnd_phdl_field_new_block(LND_PHDL_FieldTempl *templ, LND_PHDL_BitPtr *ptr);
LND_PHDL_Field         *libnd_phdl_field_new_chain(LND_PHDL_FieldTempl *templ, LND_PHDL_BitPtr *ptr);
LND_PHDL_Field         *libnd_phdl_field_new_union(LND_PHDL_FieldTempl *templ, LND_PHDL_BitPtr *ptr);
LND_PHDL_Field         *libnd_phdl_field_new_enum(LND_PHDL_FieldTempl *templ, LND_PHDL_BitPtr *ptr);

void                    libnd_phdl_field_free(LND_PHDL_Field *field);						 

LND_PHDL_Field         *libnd_phdl_field_find_field(LND_PHDL_Field *field,
						    const char *field_name);

gboolean                libnd_phdl_field_eval(LND_PHDL_Field *field, guint64 *result);

gboolean                libnd_phdl_bitptr_get_raw(LND_PHDL_BitPtr *ptr, guint len,
						  guchar *dest, guint destlen);

gboolean                libnd_phdl_bitptr_get_string_fixed(LND_PHDL_BitPtr *ptr,
							   guint length,
							   char *dest);

gboolean                libnd_phdl_bitptr_get_string_marked(LND_PHDL_BitPtr *ptr,
							    const char *marker,
							    char *dest, int destlen);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __libnd_phdl_header_h */
